// Copyright 2012 by MagicAsh and flying_galactic
// Copyright 2000 by David Brownell <dbrownell@users.sourceforge.net>
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed epIn the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
package com.android.ptplib;

import android.annotation.SuppressLint;
import android.hardware.usb.UsbConstants;
import android.hardware.usb.UsbDevice;
import android.hardware.usb.UsbDeviceConnection;
import android.hardware.usb.UsbEndpoint;
import android.hardware.usb.UsbInterface;
import android.util.Log;

/**
 * This class initiates interactions with USB devices, supporting only
 * mandatory PTP-over-USB operations; both
 * "push" and "pull" modes are supported.  Note that there are some
 * operations that are mandatory for "push" responders and not "pull"
 * ones, and vice versa.  A subclass adds additional standardized
 * operations, which some PTP devices won't support.  All low
 * level interactions with the device are done by this class,
 * including especially error recovery.
 *
 * <p> The basic sequence of operations for any PTP or ISO 15470
 * initiator (client) is:  acquire the device; wrap it with this
 * driver class (or a subclass); issue operations;
 * close device.  PTP has the notion
 * of a (single) session with the device, and until you have an open
 * session you may only invoke {@link #getDeviceInfo} and
 * {@link #openSession} operations.  Moreover, devices may be used
 * both for reading images (as from a camera) and writing them
 * (as to a digital picture frame), depending on mode support.
 *
 * <p> Note that many of the IOExceptions thrown here are actually
 * going to be <code>usb.core.PTPException</code> values.  That may
 * help your application level recovery processing.  You should
 * assume that when any IOException is thrown, your current session
 * has been terminated.
 *
 * @author MagicAsh and flying_galactic, based on JAVA code by David Brownell
 *
 * This class has been reworked by flying_galactic and MagicAsh order to make it 
 * compatible with Android. Also, this is more a derivative work than just an 
 * adaptation of the original version. 
 */
@SuppressLint("NewApi")
public class BaselineInitiator extends NameFactory implements Runnable, CameraFunctions {

	private static final byte 		CLASS_DEVICE_RESET      = (byte) 0x66;
	private static final byte 		CLASS_GET_DEVICE_STATUS = (byte) 0x67;
	protected static final int  	DEFAULT_TIMEOUT 	  = 1000; // ms
	public static final int  		BUSY_TIMEOUT_MILLIS	  = 5000; // ms
	public static final int			SLEEP_CLEAR		      = 500; // ms

	public static boolean 			DEBUG = false;
	public static boolean 			TRACE = false;
	public static final String 		TAG = "BaselineInitiator";
	public static Thread 			thread = null;

	public UsbDevice                device;
	public UsbInterface 			intf;
	public UsbEndpoint  			epIn;
	public int                    	inMaxPS;
	public UsbEndpoint  			epOut;
	public UsbEndpoint  			epEv;
	public Session                	session;
	public DeviceInfo             	info;
	public UsbDeviceConnection 		mConnection = null; // must be initialized first! 
	// mUsbManager = (UsbManager)getSystemService(Context.USB_SERVICE);



	/**
	 * Constructs a class driver object, if the device supports
	 * operations according to Annex D of the PTP specification.
	 *
	 * @param device the first PTP interface will be used
	 * @exception IllegalArgumentException if the device has no
	 *	Digital Still Imaging Class or PTP interfaces
	 */
	public BaselineInitiator(UsbDevice dev, UsbDeviceConnection connection) throws PTPException {
		if (connection == null) {
			throw new PTPException ("BaselineInitiator: Connection = null");//IllegalArgumentException();
		}
		this.mConnection = connection;
		if (dev == null) {
			throw new PTPException ("BaselineInitiator: dev = null");//IllegalArgumentException();
		}
		session = new Session();
		this.device = dev;
		intf = findUsbInterface (dev);

		if (intf == null) {
			//if (usbInterface == null) {
			throw new PTPException("BaselineInitiator: No PTP interfaces associated to the device");
		}

		for (int i = 0; i < intf.getEndpointCount(); i++) {
			UsbEndpoint ep = intf.getEndpoint(i);
			if (ep.getType() == UsbConstants.USB_ENDPOINT_XFER_BULK) {
				if (ep.getDirection() == UsbConstants.USB_DIR_OUT) {
					epOut = ep;
				} else {
					epIn = ep;
				}
			}
			if (ep.getType() == UsbConstants.USB_ENDPOINT_XFER_INT){
				epEv = ep;
			}
		}
		endpointSanityCheck();
		inMaxPS = epOut.getMaxPacketSize();

		//UsbDevice usbDevice = dev.getUsbDevice();
		//            UsbConfigDescriptor[] descriptors = usbDevice.getConfig();
		//
		//            if ((descriptors == null) || (descriptors.length < 1)) {
		//                throw new PTPException("UsbDevice with no descriptors!");
		//            }


		// clear epOut any previous state
		reset();
		if (getClearStatus() != Response.OK
				&& getDeviceStatus(null) != Response.OK) {
			throw new PTPException("BaselineInitiator: can't init");
		}

		//            UsbLogger.log (TAG, "trying getDeviceInfoUncached");
		// get info to sanity check later requests
		info = getDeviceInfoUncached(); 

		// set up to use vendor extensions, if any
		if (info.vendorExtensionId != 0) {
			info.factory = updateFactory(info.vendorExtensionId);
		}
		session.setFactory((NameFactory) this);
	}


	// searches for an interface on the given USB device, returns only class 6  // From androiddevelopers ADB-Test
	private UsbInterface findUsbInterface(UsbDevice device) {
//		UsbLogger.log  (TAG, "findUsbInterface: " + device.getDeviceName());
		int count = device.getInterfaceCount();
		for (int i = 0; i < count; i++) {
			UsbInterface intf = device.getInterface(i);
			UsbLogger.log  (TAG, "Interface " +i + " Class " +intf.getInterfaceClass() +" Prot " +intf.getInterfaceProtocol());
			if (intf.getInterfaceClass() == 6
					//255 && intf.getInterfaceSubclass() == 66 && intf.getInterfaceProtocol() == 1
					) {
				return intf;
			}
		}
		UsbLogger.log  (TAG, "findUsbInterface: no interface Class 6 found! Try interface 0");
		return null;
	}


	/**
	 * @return the device
	 */
	public UsbDevice getDevice() {
		return device;
	}

	/**
	 * Returns the last cached copy of the device info, or returns
	 * a newly cached copy.
	 * @see #getDeviceInfoUncached
	 */
	public DeviceInfo getDeviceInfo() throws PTPException {
		if (info == null) {
			return getDeviceInfoUncached();
		}
		return info;
	}

	/**
	 * Sends a USB level CLASS_DEVICE_RESET control message.
	 * All PTP-over-USB devices support this operation.
	 * This is documented to clear stalls and camera-specific suspends,
	 * flush buffers, and close the current session.
	 *
	 * <p> <em>TO BE DETERMINED:</em> How does this differ from a bulk
	 * protocol {@link Initiator#resetDevice ResetDevice} command?  That
	 * command is documented as very similar to this class operation.
	 * Ideally, only this control request will ever be used, since it
	 * works even when the bulk channels are halted.
	 */
	public void reset() throws PTPException 
	{
		//        try {
		/*
		 * JAVA: public int controlMsg(int requestType,
                      int request,
                      int value,
                      int index,
                      byte[] data,
                      int size,
                      int timeout,
                      boolean reopenOnTimeout)
               throws USBException

Android: UsbDeviceConnection controlTransfer (int requestType, int request, int value, int index, byte[] buffer, int length, int timeout)                   	
		 */		
		if (mConnection == null) throw new PTPException("No Connection");

		mConnection.controlTransfer(
				(int) ( UsbConstants.USB_DIR_OUT      |
						UsbConstants.USB_TYPE_CLASS),
                        CLASS_DEVICE_RESET,
                        0,
                        0,
                        new byte[0],
                        0,
                        DEFAULT_TIMEOUT 
				);

		session.close();
	}

	/**
	 * Issues an OpenSession command to the device; may be used
	 * with all responders.  PTP-over-USB doesn't seem to support
	 * multisession operations; you must close a session before
	 * opening a new one.
	 */
	public void openSession() throws PTPException {
		Command command;
		Response response;

		synchronized (session) {
			command = new Command(Command.OpenSession, session,
					session.getNextSessionID());
			response = transactUnsync(command, null);
			switch (response.getCode()) {
			case Response.OK:
				session.open();
				thread = new Thread (this);
				thread.start(); // Starts Event listener
				return;
			default:
				throw new PTPException(response.toString());
			}
		}
	}

	/**
	 * Issues a CloseSession command to the device; may be used
	 * with all responders.
	 */
	public void closeSession() throws PTPException {
		Response response;

		synchronized (session) {
			// checks for session already open
			response = transact0(Command.CloseSession, null);
			switch (response.getCode()) {
			case Response.SessionNotOpen:
				if (DEBUG) {
					System.err.println("close unopen session?");
				}
				// FALLTHROUGH
			case Response.OK:
				session.close();
				return;
			default:
				throw new PTPException(response.toString());
			}
		}
	}

	/**
	 * Closes the session (if active) and releases the device.
	 *
	 * @throws PTPException
	 */
	public void close() throws PTPException {
		if (isSessionActive()) {
			try {
				closeSession();
			} catch (PTPException ignore) {
				// If we cannot close the session, there is nothing we can do
			}
		}

		try {
			if (mConnection != null && intf != null) mConnection.releaseInterface(intf);
			if (mConnection != null) mConnection.close();
			device = null;
			info = null;
		} catch (Exception ignore) {
			throw new PTPException("Unable to close the USB device");
		}
	}

	public Response showResponse (Response response) {
		UsbLogger.log (TAG, "  Type: " + Command.getBlockTypeName(response.getBlockType()) +" (Code: " +response.getBlockType() +")\n");   // getU16 (4)
		UsbLogger.log (TAG, "  Name: " + response.getCodeName(response.getCode())+ ", code: 0x" +Integer.toHexString(response.getCode()) +"\n"); //getU16 (6)
		//			log ("  CodeString:" + response.getCodeString()+ "\n");
		UsbLogger.log (TAG, "  Length: " + response.getLength()+ " bytes\n");  //getS32 (0)
		UsbLogger.log (TAG, "  String: " + response.toString());
		return response;
	}


	public void showResponseCode (String comment, int code){
		UsbLogger.log (TAG, comment +" Response: " +getResponseString(code));
	}

	/**
	 * @return true if the current session is active, false otherwise
	 */
	public boolean isSessionActive() {
		synchronized (session) {
			return session.isActive();
		}
	}

	///////////////////////////////////////////////////////////////////
	// mandatory for all responders:  generating events
	/**
	 * Makes the invoking Thread read and report events reported
	 * by the PTP responder, until the Initiator is closed.
	 */
	@Override
	public void run() {

	}

	// ------------------------------------------------------- Protected methods

	///////////////////////////////////////////////////////////////////
	/**
	 * Performs a PTP transaction, passing zero command parameters.
	 * @param code the command code
	 * @param data data to be sent or received; or null
	 * @return response; check for code Response.OK before using
	 *	any positional response parameters
	 */
	public Response transact0(int code, Data data)
			throws PTPException {
		synchronized (session) {
			Command command = new Command(code, session);
			return transactUnsync(command, data);
		}
	}

	/**
	 * Performs a PTP transaction, passing one command parameter.
	 * @param code the command code
	 * @param data data to be sent or received; or null
	 * @param p1 the first positional parameter
	 * @return response; check for code Response.OK before using
	 *	any positional response parameters
	 */
	public Response transact1(int code, Data data, int p1)
			throws PTPException {
		synchronized (session) {
			Command command = new Command(code, session, p1);
			return transactUnsync(command, data);
		}
	}

	/**
	 * Performs a PTP transaction, passing two command parameters.
	 * @param code the command code
	 * @param data data to be sent or received; or null
	 * @param p1 the first positional parameter
	 * @param p2 the second positional parameter
	 * @return response; check for code Response.OK before using
	 *	any positional response parameters
	 */
	public Response transact2(int code, Data data, int p1, int p2)
			throws PTPException {
		synchronized (session) {
			Command command = new Command(code, session, p1, p2);
			return transactUnsync(command, data);
		}
	}

	/**
	 * Performs a PTP transaction, passing three command parameters.
	 * @param code the command code
	 * @param data data to be sent or received; or null
	 * @param p1 the first positional parameter
	 * @param p2 the second positional parameter
	 * @param p3 the third positional parameter
	 * @return response; check for code Response.OK before using
	 *	any positional response parameters
	 */
	public Response transact3(int code, Data data, int p1, int p2, int p3)
			throws PTPException {
		synchronized (session) {
			Command command = new Command(code, session, p1, p2, p3);
			return transactUnsync(command, data);
		}
	}

	// like getDeviceStatus(),
	// but clears stalled endpoints before returning  (not working under Android?!?)
	// (except when exceptions are thrown)
	// returns -1 if device wouldn't return OK status
	public int getClearStatus() throws PTPException {
//		byte [] byteBuffer = new byte [3*4];
//		Buffer buf = new Buffer(byteBuffer, byteBuffer.length);
		Buffer buf = new Buffer(null, 0);
//		UsbLogger.log (TAG, "getClearStatus");
		int retval = getDeviceStatus(buf);

		// any halted endpoints to clear?  (always both)
		if (buf.length != 4) {
			while ((buf.offset + 4) <= buf.length) {
				int ep = buf.nextS32();
				if (epIn.getAddress() == ep) {
					if (TRACE) {
						UsbLogger.log (TAG, "clearHalt epIn");
					}
					clearHalt (); //(epIn);
				} else if (epOut.getAddress() == ep) {
					if (TRACE) {
						UsbLogger.log (TAG, "clearHalt epOut");
					}
					clearHalt(); //(epOut);
				} else {
					if (DEBUG || TRACE) {
						UsbLogger.log (TAG, "?? halted EP: " + ep);
					}
				}
			}

			// device must say it's ready
			int status = Response.Undefined;

			for (int i = 0; i < 10; i++) {
				try {
					status = getDeviceStatus(null);
				} catch (PTPException x) {
					if (DEBUG) {
						x.printStackTrace();
					}
				}
				if (status == Response.OK) {
					break;
				}
				if (TRACE) {
					UsbLogger.log (TAG, "sleep; status = "
							+ getResponseString(status));
				}

				try {
					Thread.sleep(SLEEP_CLEAR);
				} catch (InterruptedException x) {
				}
			}
			if (status != Response.OK) {
				retval = -1;
			}
		} else {
			if (TRACE) {
				UsbLogger.log (TAG, "no endpoints halted");
			}
		}
		return retval;
	}

	// returns Response.OK, Response.DeviceBusy, etc
	// per fig D.6, response may hold stalled endpoint numbers
	private int getDeviceStatus(Buffer buf)
			throws PTPException {
		if (mConnection == null) throw new PTPException("No Connection");

		byte[] data = new byte[33];

		mConnection.controlTransfer(
				(int) (UsbConstants.USB_DIR_IN        |
						UsbConstants.USB_TYPE_CLASS),
                		CLASS_GET_DEVICE_STATUS,
                		0,
                		0,
                		data,
                		data.length, // force short reads
                		DEFAULT_TIMEOUT
				);

		if (buf == null) {
			buf = new Buffer(data);
		} else {
			buf.data = data;
		}
		buf.offset = 4;
		buf.length = buf.getU16(0);
//		UsbLogger.log (TAG, "getDeviceStatus: " +Response.getResponseString(buf.getU16(2)));
//		buf.dump();
//		UsbLogger.log (TAG, "getDeviceStatus: buf.length: " +buf.length +",  buf.data.length: " +buf.data.length);

		// Nikon D3000 will return only 4 Byte length!
//		if (buf.length != buf.data.length) {			
		if (buf.length < 4) {
			throw new PTPException("DeviceStatus error, Buffer length wrong!");
		}
		return buf.getU16(2);
	}

	// add event listener
	// rm event listener
	///////////////////////////////////////////////////////////////////
	// mandatory for all responders
	/**
	 * Issues a GetDeviceInfo command to the device; may be used
	 * with all responders.  This is the only generic PTP command
	 * that may be issued both inside or outside of a session.
	 */
	private DeviceInfo getDeviceInfoUncached()
			throws PTPException {
		DeviceInfo data = new DeviceInfo(this);
		Response response= null;

		synchronized (session) 
		{
//			UsbLogger.log (TAG, "getDeviceInfoUncached");
			Command command;
			try{
				command = new Command(Command.GetDeviceInfo, session);
//			try {
				response = transactUnsync(command, data);
			} catch (Exception e){
				e.printStackTrace();
				if (response == null) throw new PTPException("PTP_Error_1");
			}
//				try {
//					thread.sleep(500);
//				} catch (InterruptedException e1) {
//					// TODO Auto-generated catch block
//					e1.printStackTrace();
//				}
//				response = transactUnsync(command, data);
//			}
			UsbLogger.log (TAG, "getDeviceInfoUncached finished, " +Response._getResponseString(response.getCode()) +" responsecode: " +response.getCode());             
		}

		switch (response.getCode()) {
		case Response.OK:
			info = data;
			return data;
		default:
			if (response == null) {
				throw new PTPException("PTP_Error_2");
			} else throw new PTPException(response.toString());
		}
	}


	//TODO
	///////////////////////////////////////////////////////////////////
	// INVARIANTS:
	// - caller is synchronized on session
	// - on return, device is always epIn idle/"command ready" state
	// - on return, session was only closed by CloseSession
	// - on PTPException, device (and session!) has been reset
	public Response transactUnsync(Command command, Data data)
			throws PTPException {
		if (!"command".equals(Container.getBlockTypeName(command.getBlockType()))) {
			throw new IllegalArgumentException(command.toString());
		}
		// UsbLogger.log (TAG, command.toString() + "   Data: " +data.toString());

		// sanity checking
		int opcode = command.getCode();

		if (session.isActive()) {
			if (Command.OpenSession == opcode) {
				throw new IllegalStateException("session already open");
			}
		} else {
			if (Command.GetDeviceInfo != opcode
					&& Command.OpenSession != opcode) {
				throw new IllegalStateException("no session");
			}
		}

		// this would be UnsupportedOperationException ...
		// except that it's not available on jdk 1.1
		if (info != null && !info.supportsOperation(opcode)) {
			throw new UnsupportedOperationException(command.getCodeName(opcode));
		}

		// ok, then we'll really talk to the device
		Response response;

		// issue command
		// rejected commands will stall both EPs
		if (TRACE || DEBUG) {
			UsbLogger.logInFile ("Command: " +command.toString());
		}
		mConnection.bulkTransfer(epOut, command.data , command.length , DEFAULT_TIMEOUT);

		// may need to terminate request with zero length packet
		if ((command.length % epOut.getMaxPacketSize()) == 0) {
			mConnection.bulkTransfer(epOut, command.data, 0, DEFAULT_TIMEOUT);
		}

		// data exchanged?
		// errors or cancel (another thread) will stall both EPs
		if (data != null) {

			// write data?
			if (!data.isIn()) {
				//                	UsbLogger.log (TAG, "Start Write Data");

				data.offset = 0;
				data.putHeader(data.getLength(), Container.BLOCK_TYPE_DATA, opcode, command.getXID());

//				if (TRACE) {
					UsbLogger.logInFile ("Data: " +data.toString());
//				}

				// Special handling for the read-from-N-mbytes-file case
				//TODO yet to be implemented
				//                    if (data instanceof FileSendData) {
				//                        FileSendData fd = (FileSendData) data;
				//                        int len = fd.data.length - fd.offset;
				//                        int temp;
				//
				//                        // fill up the rest of the first buffer
				//                        len = fd.read(fd.data, fd.offset, len);
				//                        if (len < 0) {
				//                            throw new PTPException("eh? " + len);
				//                        }
				//                        len += fd.offset;
				//
				//                        for (;;) {
				//                            // write data or terminating packet
				//                        	mConnection.bulkTransfer(epOut, data , command.length , DEFAULT_TIMEOUT);
				//                            //stream.write(fd.data, 0, len);
				//                            if (len != fd.data.length) {
				//                                break;
				//                            }
				//
				//                            len = fd.read(fd.data, 0, fd.data.length);
				//                            if (len < 0) {
				//                                throw new PTPException("short: " + len);
				//                            }
				//                        }
				//
				//                    } else {
				// write data and maybe terminating packet
				byte[] bytes = data.getData();//new byte [data.length];
				UsbLogger.logInFile (TAG, "   send Data, " +data.toString() +String.format(" Code: 0x%04X", data.getCode()));
				data.dump();
				int len = mConnection.bulkTransfer(epOut, bytes , bytes.length, DEFAULT_TIMEOUT);
				UsbLogger.log (TAG, "   Data bytes sent: " +len);
				if (len < 0) {
					throw new PTPException("short: " + len);
				}

				if ((data.length % epOut.getMaxPacketSize()) == 0) {
					UsbLogger.log (TAG, "send 0 Data");
					mConnection.bulkTransfer(epOut, bytes , 0, DEFAULT_TIMEOUT);
					//stream.write(data.data, 0, 0);
					//                        }
				}

				// read data?
			} else {
				//                	UsbLogger.log (TAG, "Start Read Data");
				byte buf1[] = new byte[inMaxPS];
				int len = 0;
//				len = device.getInputStream(epIn).read(buf1);
//				UsbLogger.logInFile (TAG, "Receive data, max " +inMaxPS);
				len = mConnection.bulkTransfer(epIn, buf1, inMaxPS , DEFAULT_TIMEOUT);
				if (len > 12 || len < 0)   UsbLogger.logInFile  (TAG, "received data bytes: " +len +", inMaxPS: " +inMaxPS);
				if (len < 0) throw new PTPException("protocol err, no data received! len: " +len); // Canon 6D returns -1 on connect!

				// Get the first bulk packet(s), check header for length
				data.data = buf1;
				data.length = len;
				//                    if (len > 16) data.dump(); 
//				if (TRACE) {
					UsbLogger.logInFile ("Data: " +data.toString());
//				}
				if (data.getBlockType()   != Container.BLOCK_TYPE_DATA
						|| data.getCode() != command.getCode()
						|| data.getXID()  != command.getXID()
						) {
					UsbLogger.log (TAG, "PTPError, BlockType (should be DATA): " +Container.getBlockTypeName(data.getBlockType()));
					UsbLogger.log (TAG, "          data.code = command.code: " +(data.getCode() == command.getCode()));
					UsbLogger.log (TAG, "          data.XID  = command.XID:  " +(data.getXID()  == command.getXID()));
					data.dump();
					throw new PTPException("protocol err 1, " + data);
				}

				// get the rest of it
				int expected = data.getLength();

				// Special handling for the write-to-N-mbytes-file case
				//TODO yet to be implemented
				//                    if (data instanceof OutputStreamData) {
				//                        OutputStreamData fd = (OutputStreamData) data;
				//                        fd.write(buf1, Data.HDR_LEN, len - Data.HDR_LEN);
				//                        if (len == inMaxPS && expected != inMaxPS) {
				//                            InputStream is = device.getInputStream(epIn);
				//
				//                            // at max usb data rate, 128K ~= 0.11 seconds
				//                            // typically it's more time than that
				//                            buf1 = new byte[128 * 1024];
				//                            do {
				//                                len = is.read(buf1);
				//                                fd.write(buf1, 0, len);
				//                            } while (len == buf1.length);
				//                        }
				//                    } else 
				if (len == inMaxPS && expected != inMaxPS) {
					buf1 = new byte[expected];
					System.arraycopy(data.data, 0, buf1, 0, len);
					data.data = buf1;
					UsbLogger.log (TAG, "read additional bytes " +(expected - len));
					data.length += mConnection.bulkTransfer(epIn, buf1, expected - len, DEFAULT_TIMEOUT);//device.getInputStream(epIn).read(buf1, len, expected - len);
					UsbLogger.log (TAG, "Read oK");
				}

				// if ((expected % inMaxPS) == 0)
				//	... next packet will be zero length

				// and do whatever parsing needs to be done
				try {
					data.parse();
				} catch (Exception e) {
					UsbLogger.log (TAG, e.toString() +"\n in " +e.getStackTrace()[0].getClassName() +"  " +e.getStackTrace()[0].getMethodName() +" line " + e.getStackTrace()[0].getLineNumber());		
				}
			}
		}

		// (short) read the response
		// this won't stall anything
		byte buf[] = new byte[inMaxPS];
		            UsbLogger.logInFile (TAG, "read response");
		int len = mConnection.bulkTransfer(epIn, buf ,inMaxPS , DEFAULT_TIMEOUT);//device.getInputStream(epIn).read(buf);
		            if (len != 12) UsbLogger.log (TAG, "1. received data bytes: " +len);

		// ZLP terminated previous data?
		if (len <= 0) { // < 0 = error
			len = mConnection.bulkTransfer(epIn, buf ,inMaxPS , DEFAULT_TIMEOUT);// device.getInputStream(epIn).read(buf);
			                UsbLogger.logInFile (TAG, "2. received data bytes: " +len);
		}
		if (len < 0) {
			            	UsbLogger.logInFile (TAG, "Data receive error, len < 0: " +len);
			response = new Response(new byte[Response.HDR_LEN], this);
			response.putHeader(Response.HDR_LEN, 3 /*response*/, Response.GeneralError, command.getXID());
		} else { 
			response = new Response(buf, len, this);
			UsbLogger.logInFile ("Response: ", buf, len);
//			UsbLogger.logInFile (response.toString());
		}

//		if (TRACE) {
//		    UsbLogger.logInFile ("Command: "  +command.toString());
//		    UsbLogger.logInFile ("Data: "     +data.toString());
			UsbLogger.logInFile ("Response: " +response.toString());
//		}

		return response;

		//TODO implement stall detection
		//       } catch (USBException e) {
		//            if (DEBUG) {
		//                e.printStackTrace();
		//            }
		//
		//            // PTP devices will stall bulk EPs on error ... recover.
		//            if (e.isStalled()) {
		//                int status = -1;
		//
		//                try {
		//                    // NOTE:  this is the request's response code!  It can't
		//                    // be gotten otherwise; despite current specs, this is a
		//                    // "control-and-bulk" protocol, NOT "bulk-only"; or more
		//                    // structurally, the protocol handles certain operations
		//                    // concurrently.
		//                    status = getClearStatus();
		//                } catch (PTPException x) {
		//                    if (DEBUG) {
		//                        x.printStackTrace();
		//                    }
		//                }
		//
		//                // something's very broken
		//                if (status == Response.OK || status == -1) {
		//                    throw new PTPException(e.getMessage(), e);
		//                }
		//
		//                // treat status code as the device's response
		//                response = new Response(new byte[Response.HDR_LEN], this);
		//                response.putHeader(Response.HDR_LEN, 3 /*response*/,
		//                        status, command.getXID());
		//                if (TRACE) {
		//                    System.err.println("STALLED: " + response.toString());
		//                }
		//
		//                abort = false;
		//                return response;
		//            }
		//            throw new PTPException(e.getMessage(), e);
		//
		//        } catch (IOException e) {
		//            throw new PTPException(e.getMessage(), e);
		//
		//        } finally {
		//            if (abort) {
		//                // not an error we know how to recover;
		//                // bye bye session!
		//                reset();
		//            }
		//        }
	}

	private void endpointSanityCheck() throws PTPException {
		if (epIn == null) {
			throw new PTPException("No input end-point found!");
		}

		if (epOut == null) {
			throw new PTPException("No output end-point found!");
		}

		if (epEv == null) {
			throw new PTPException("No input interrupt end-point found!");
		}
		if (DEBUG){
			UsbLogger.log (TAG, "Get: "+device.getInterfaceCount()+" Other: "+device.getDeviceName());
			UsbLogger.log (TAG, "\nClass: "+intf.getInterfaceClass()+","+intf.getInterfaceSubclass()+","+intf.getInterfaceProtocol()
					+ "\nIendpoints: "+epIn.getMaxPacketSize()+ " Type "+ epIn.getType() + " Dir "+epIn.getDirection()); //512 2 USB_ENDPOINT_XFER_BULK USB_DIR_IN 
			UsbLogger.log (TAG, "\nOendpoints: "+epOut.getMaxPacketSize()+ " Type "+ epOut.getType() + " Dir "+epOut.getDirection()); //512 2 USB_ENDPOINT_XFER_BULK USB_DIR_OUT 
			UsbLogger.log (TAG, "\nEendpoints: "+epEv.getMaxPacketSize()+ " Type "+ epEv.getType() + " Dir "+epEv.getDirection()); //8,3 USB_ENDPOINT_XFER_INT USB_DIR_IN  

		}
	}

	private void clearHalt (){//  (UsbEndpoint e) {
//		if (e.getDirection() == UsbConstants.USB_DIR_IN) {
//                    InputStream in;
//                    in = e.
//                    // Read in data here
//                    in.close();
//                }
//                else {
//                    // If the Endpoint is an output Endpoint,
//                    // obtain its OutputStream and write out data.
//                    OutputStream out;
//                    out = ep.getOutputStream();
//                    out.close();
//                }

		//
		// TODO: implement clearHalt of an endpoint
		//
		mConnection.releaseInterface(intf); //!? stackoverflow.com/questions/9574197/is-there-a-way-to-reset-the-usb-bulk-endpoints-using-the-android-usb-host-apis
	}

	public static final int STD_BUSY_TIMEOUT = 10000;

	// initiates a capture by calling the overridden procedure and waits undtil not busy
	public Response initiateCapture(int storageId, int formatCode, int busyTimeoutMillis)
			throws PTPException {
		Response response;
		long t0 = System.currentTimeMillis();
		do {
			response = initiateCapture(storageId, formatCode);
			if (response.getCode() == Response.DeviceBusy) {
				UsbLogger.log (TAG, "initiateCapture: Device busy, Waiting...");
				try {
					Thread.sleep (200);
				} catch (InterruptedException e) {
					e.printStackTrace();
				}
			}
		} while (response.getCode() == Response.DeviceBusy 
				&& (System.currentTimeMillis() -t0 < busyTimeoutMillis));
		return response;
	}

	public Response bulbExposure (int milliseconds)  throws PTPException{
		Response resp = null;
		resp = startBulb ();
		try {
			Thread.sleep(milliseconds);
		} catch (InterruptedException e) {e.printStackTrace();}

		resp = stopBulb ();
		//			showResponseCode ("  EosBulbEnd", resp.getCode());
		return resp;
	}

	public Response setShutterSpeed (double timeSeconds, int busyTimeoutMillis) throws PTPException{
		Response response;
		long t0 = System.currentTimeMillis();
		do {
			response = setShutterSpeed (timeSeconds);
			if (response.getCode() == Response.DeviceBusy) {
				UsbLogger.log (TAG, " Device busy, Waiting...");
				try {
					Thread.sleep (100);
				} catch (InterruptedException e) {
					e.printStackTrace();
				}
			}
		} while (response.getCode() == Response.DeviceBusy 
				&& (System.currentTimeMillis() -t0 < busyTimeoutMillis));
		return response;
	}

	public Response setAperture(double apertureValue, int busyTimeoutMillis) throws PTPException{
		Response response;
		long t0 = System.currentTimeMillis();
		do {
			response = setAperture(apertureValue);
			if (response.getCode() == Response.DeviceBusy) {
				UsbLogger.log (TAG, " Device busy, Waiting...");
				try {
					Thread.sleep (100);
				} catch (InterruptedException e) {
					e.printStackTrace();
				}
			}
		} while (response.getCode() == Response.DeviceBusy 
				&& (System.currentTimeMillis() -t0 < busyTimeoutMillis));
		return response;
	}

	// Sets ISO to 50, 100, 200... or nearest value
	public Response setISO(double isoValue, int busyTimeoutMillis) throws PTPException{
		Response response;
		long t0 = System.currentTimeMillis();
		do {
			response = setISO(isoValue);
			if (response.getCode() == Response.DeviceBusy) {
				UsbLogger.log (TAG, " Device busy, Waiting...");
				try {
					Thread.sleep (100);
				} catch (InterruptedException e) {
					e.printStackTrace();
				}
			}
		} while (response.getCode() == Response.DeviceBusy 
				&& (System.currentTimeMillis() -t0 < busyTimeoutMillis));
		return response;
	}

	public Response setExposure(double exposureValue, int busyTimeoutMillis) throws PTPException{  
		Response response;
		long t0 = System.currentTimeMillis();
		do {
			response = setExposure(exposureValue);
			if (response.getCode() == Response.DeviceBusy) {
				UsbLogger.log (TAG, " Device busy, Waiting...");
				try {
					Thread.sleep (100);
				} catch (InterruptedException e) {
					e.printStackTrace();
				}
			}
		} while (response.getCode() == Response.DeviceBusy 
				&& (System.currentTimeMillis() -t0 < busyTimeoutMillis));
		return response;
	}

	public Response setImageQuality (String quality, int busyTimeoutMillis) throws PTPException{
		Response response;
		long t0 = System.currentTimeMillis();
		do {
			response = setImageQuality(quality);
			if (response.getCode() == Response.DeviceBusy) {
				UsbLogger.log (TAG, " Device busy, Waiting...");
				try {
					Thread.sleep (100);
				} catch (InterruptedException e) {
					e.printStackTrace();
				}
			}
		} while (response.getCode() == Response.DeviceBusy 
				&& (System.currentTimeMillis() -t0 < busyTimeoutMillis));
		return response;
	}


	/*************************************************************************************
	 *                                                                                   *
	 *	  Methods to be overridden in camera-specific instances of baselineInititator    *
	 *                                                                                   *
	 *    the following methods are like abstract methods to be declared                 *
	 *    properly in the appropriate device initiators                                  *
	 *                                                                                   *
	 *************************************************************************************/

	 
	public Response initiateCapture(int storageId, int formatCode)
			throws PTPException {
		return null;
	}
	public Response startBulb () throws PTPException{
		return null;
	}

	public Response stopBulb () throws PTPException{
		return null;
	}
	public Response setShutterSpeed (int speed) throws PTPException{
		return null;
	}  

	public Response setExposure(int exposure) throws PTPException{    	
		return null;
	}

	public Response setISO(int value) throws PTPException{
		return null;
	}

	public Response setAperture(int value) throws PTPException{
		return null;
	}

	public Response setImageQuality(int value) throws PTPException{
		return null;
	}

	// Floating point adapters
	public Response setShutterSpeed (double timeSeconds) throws PTPException{
		return null;
	}

	public Response setAperture(double apertureValue) throws PTPException{
		return null;
	}

	// Sets ISO to 50, 100, 200... or nearest value
	public Response setISO(double isoValue) throws PTPException{
		return null;
	}

	public Response setExposure(double exposureValue) throws PTPException{  
		UsbLogger.log (TAG, "Not overriden!!!");
		return null;
	}

	/**  Selects image Quality from "S" to "RAW" in 4 steps
	 * 
	 */
	public Response setImageQuality (String quality) throws PTPException{
		return null;
	}


	@Override
	public Response getDevicePropValue(int property, DevicePropDesc desc)
			throws PTPException {
		// TODO Auto-generated method stub
		return null;
	}


}
