// Copyright 2000 by David Brownell <dbrownell@users.sourceforge.net>
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
package com.android.ptplib.eos;

import java.io.ByteArrayInputStream;
import java.util.ArrayList;
import java.util.List;

import android.annotation.SuppressLint;
import android.hardware.usb.UsbDevice;
import android.hardware.usb.UsbDeviceConnection;
import android.util.Log;


import com.android.ptplib.BaselineInitiator;
import com.android.ptplib.CameraFunctions;
import com.android.ptplib.Command;
import com.android.ptplib.Container;
import com.android.ptplib.Data;
import com.android.ptplib.DevicePropDesc;
import com.android.ptplib.PTPException;
import com.android.ptplib.PTPUnsupportedException;
import com.android.ptplib.Response;

/**
 * This supports all standardized PTP-over-USB operations, including
 * operations (and modes) that are optional for all responders.
 * Filtering operations invoked on this class may be done on the device,
 * or may be emulated on the client side.
 * At this time, not all standardized operations are supported.
 *
 * @version $Id: EosInitiator.java,v 1.9 2001/04/12 23:13:00 dbrownell Exp $
 * @author David Brownell
 */
@SuppressLint("NewApi")
public class EosInitiator extends BaselineInitiator implements EosEventConstants, CameraFunctions{
	
	public static final int CANON_VID = 1193; // 0x04a9;  see http://www.linux-usb.org/usb.ids
	
	public static boolean eventListenerRunning = false;

    /**
     * This is essentially a class driver, following Annex D of
     * the PTP specification.
     */
    @SuppressLint("NewApi")
	public EosInitiator(UsbDevice dev, UsbDeviceConnection connection) throws PTPException {
        super(dev, connection);
        if (thread != null) eventListenerRunning = false;
        thread = new Thread (this);
        thread.start();
    }

    /**  Does not work, returns PTPerror 
     * Fills out the provided device property description.
     *
     * @param propcode code identifying the property of interest
     * @param desc description to be filled; it may be a subtype
     *	associated with with domain-specific methods
     * @return response code
     */
    public int getDevicePropDesc(int propcode, DevicePropDesc desc)
            throws PTPException {
        return transact1(Command.GetDevicePropDesc, desc, propcode).getCode();
    }

	private byte[] read(int timeout)
	{
		Log.d(TAG,"Reading data");
		byte data[] = new byte[inMaxPS]; 
		int lengthOfBytes = mConnection.bulkTransfer(epIn, data, inMaxPS, timeout);	  
		return data;	  
	}

    @Override
    public Response getDevicePropValue(int property, DevicePropDesc desc)
    		throws PTPException {
		Command command = new Command(Command.GetDevicePropDesc, session, property);
		mConnection.bulkTransfer(epOut, command.data, command.length, BaselineInitiator.DEFAULT_TIMEOUT);    
		byte buf[] = read(BaselineInitiator.DEFAULT_TIMEOUT);

		desc.data = buf;
		desc.length = info.getLength();
		desc.parse(); // crashes on 2nd call without reading response!
		
		// do not forget, read response otherwise may crash!		
        buf = new byte[inMaxPS];
        int len = mConnection.bulkTransfer(epIn, buf ,inMaxPS , DEFAULT_TIMEOUT);
        return new Response(buf, len, this);
	}

    
    //TODO set this to private
    public  Response setDevicePropValueEx   (int property, int value) throws PTPException{
    	return setDevicePropValue (property, value);
    }
    
	public Response setDevicePropValue (int property, int value) throws PTPException{
    	byte [] buff = new byte [0x18];
    	Data data = new Data (false, buff, buff.length, this);
        data.offset = 0;
        data.putHeader(buff.length, Container.BLOCK_TYPE_DATA, Command.EosSetDevicePropValueEx, 0 /*XID, dummy, will be overwritten*/);
    	data.put32 (0x0c); // Length: 12 bytes
    	data.put32 (property);
    	data.put32 (value);	
		return transact0(Command.EosSetDevicePropValueEx, data); 
	}


	
    /**
     * Checks if there is any event available. If there is any, an ArrayList
     * of Event object is returned.
     *
     * @return the list of available events
     *
     * @throws PTPException in case of errors
     */
    public List<EosEvent> checkEvents()
            throws PTPException {
    	int ret;
//        ret = transact1(Command.EosSetEventMode, null, 1).getCode(); // Prevents Releasing shutter
//        showResponseCode ("  EosSetEventMode 1 ", ret);
//        if (ret != Response.OK) {     	
//            throw new PTPException("Error reading events", ret);
//        }

    	
    	// Synchronous transaction
        Data data = new Data(this);
        Response res = transact0(Command.EosGetEvent, data);
        if (res.getLength() > 12  &&/* must be at least 20 bytes only confirmation */
        		res.getCode() == Response.OK)
        // Log to LogCat	
        showResponseCode ("  GetEvent: data length: " +data.getLength() +" resLength: " +res.getLength() +" EosGetEvent ", res.getCode());
        
        if (res.getCode() != Response.OK) {
            throw new PTPException(String.format("Failed getting events from the camera (%1$04X)", res.getCode()));
        }

        if (data.getLength() < 12) { // catch error data too small
            throw new PTPException(String.format("Failed getting events from the camera (data.length: %d)", data.getLength()));
        }
        //
        // We need to discard the initial 12 USB header bytes
        //
        byte[] buf = new byte[data.getLength() - 12];
        System.arraycopy(data.getData(), 12, buf, 0, buf.length);

        EosEventParser parser = new EosEventParser(new ByteArrayInputStream(buf));

        ArrayList<EosEvent> events = new ArrayList<EosEvent>();
        while (parser.hasEvents()) {
            try {
            	EosEvent event = parser.getNextEvent();
            	events.add(event);
            } catch (PTPUnsupportedException e) {
                //
                // TODO: log this information?
                //
                //System.err.println("Skipping unsupported event");
            }
        }

        return events;
    }

        
    /**
     * Starts the capture of one (or more) new
     * data objects, according to current device properties.
     * The capture will complete without issuing further commands.
     *
     * @see #initiateOpenCapture
     *
     * @param storageId Where to store the object(s), or zero to
     *	let the device choose.
     * @param formatCode Type of object(s) to capture, or zero to
     *	use the device default.
     *
     * @return status code indicating whether capture started;
     *	CaptureComplete events provide capture status, and
     *	ObjectAdded events provide per-object status.
     */
    @Override
    public Response initiateCapture(int storageId, int formatCode)
            throws PTPException {
    	Response resp = null;
        //
        // Special initialization for EOS cameras
        //
        if (!info.supportsOperation(Command.EosRemoteRelease)) {
        	Log.d(TAG, "The camera does not support EOS capture");
            throw new PTPException("The camera does not support EOS capture");
        }

        resp = transact0(Command.EosRemoteRelease, null);
        int ret = resp.getCode();
        //Log.d(TAG, "  EosRemoteRelease " +Response.getCodeString(ret) +" code: 0x" +Integer.toHexString(ret) +"  OK: " +(ret == Response.OK));
        if (ret != Response.OK && ret != Response.DeviceBusy) {
            String msg = "Canon EOS Capture failed to release: Unknown error "
                    + ret
                    + " , please report.";
            if (ret == 1) {
                msg = "Canon EOS Capture failed to release: Perhaps no focus?";
            } else if (ret == 7) {
                msg = "Canon EOS Capture failed to release: Perhaps no more memory on card?";
            }
            Log.d(TAG, msg);
            throw new PTPException(msg, ret);
        }
        return resp;
    }

    /**
     * Retrieves a chunk of the object identified by the given object id.
     *
     * @param oid object id
     * @param offset the offset to start from
     * @param size the number of bytes to transfer
     * @param data the Data object receiving the object
     *
     * @throws PTPException in case of errors
     */
    public void getPartialObject(int oid, int offset, int size, Data data)
    throws PTPException {
        Response ret =
            transact3(Command.EosGetPartialObject, data, oid, offset, size);

        if (ret.getCode() != Response.OK) {
            throw new PTPException("Error reading new object", ret.getCode());
        }
    }

    public void transferComplete(int oid)
    throws PTPException {
        Response ret =
            transact1(Command.EosTransferComplete, null, oid);

        if (ret.getCode() != Response.OK) {
            throw new PTPException("Error reading new object", ret.getCode());
        }
    }
    
    @Override 
    public void closeSession() throws PTPException {
    	eventListenerRunning = false;
    	super.closeSession();
    }
    
    
    @Override    
    public void run() {
    	if (eventListenerRunning) return; //already running?
    	Log.d(TAG, "*** EOSInitiator.run : Start Event listener");
    	eventListenerRunning = true;
    	if (session != null && isSessionActive()) {
    		Response resp;
			try {
				resp = transact1(Command.EosSetEventMode, null, 1);
//	        	showResponseCode ("  EosSetEventMode 1: ", resp.getCode());
			} catch (PTPException e) {e.printStackTrace(); }
    	}
    	while (eventListenerRunning) {
    		try {
    			if (session != null && isSessionActive()) {
    				checkEvents();
   				
    		    	// Asynchronous transaction
//    		    	final int BUFLEN = 12;
//    		        ByteBuffer buffer = ByteBuffer.allocate(BUFLEN);
//    		        UsbRequest request = new UsbRequest();
//    		        request.initialize(mConnection, epEv);
//    		        while (session.isActive()) {
//    		            // queue a request on the interrupt endpoint
//    		            request.queue(buffer, BUFLEN);
//    		            // send poll status command
////    		            sendCommand(COMMAND_STATUS);
//    		            // wait for status event
//    		            if (mConnection.requestWait() == request) {
//    		                byte buf[] = buffer.array();
//    		                EosEvent event = new EosEvent ();
//    		        		if (event.getCode() != 0) {
//    		        			Log.d(TAG, "Event received");
////    		        			Log.d(TAG, "  Event Name: " + event.getCodeName(event.getCode())+ ", code: 0x" +Integer.toHexString(event.getCode()));
////    		        			Log.d(TAG, "  Length:" + event.getLength()+ "\n");
////    		        			Log.d(TAG, "  String:" + event.toString() +"\n");
//    		        		}
//    		                try {
//    		                    Thread.sleep(100);
//    		                } catch (InterruptedException e) {
//    		                }
//    		            } else {
//    		                Log.e(TAG, "requestWait failed, exiting");
//    		                break;
//    		            }
//    		        }
    			}
    		} catch (PTPException e1) {
    			//e1.printStackTrace();
    		} 
    		try { Thread.sleep(100);  } catch (InterruptedException e) {e.printStackTrace();}    
    	}
    	eventListenerRunning = false;
    }
    
    /** Sets EOS Image Quality
     * 
     * @param size S = 2, M = 1, L = 0
     * @param raw
     * @param fine
     * @param formatsToSave 2 = Raw +L, 1 = single image
     * @return
     * @throws PTPException
     */
	public Response setEOSImgQual (int size, boolean raw, boolean fine, int formatsToSave) throws PTPException{
		int blockLength = 0x0c +16* formatsToSave;
		byte [] buff = new byte [blockLength+12]; //Total length
    	Data data = new Data (false, buff, buff.length, this);
        data.offset = 0;
        data.putHeader(buff.length, Container.BLOCK_TYPE_DATA, Command.EosSetDevicePropValueEx, 0 /*XID, dummy, will be overwritten*/);
    	data.put32 (blockLength); // Length of formats  28d = 0x1C
    	data.put32 (Command.EOS_DPC_ImageQuality);
    	data.put32 (formatsToSave);	// images to store, 2 = Raw +L, 1 = single image	        	
    	data.put32 (0x10);	// block length info	  
    	
    	if (raw) {data.put32 (0x06);} else data.put32 (0x01);	// RAW = 6, JPG = 1	        	
    	data.put32 (size);	// S = 2, M = 1, L = 0, S1 = 0x0e, S2 = 0x0f, S3 = 0x10
    	if (raw) {data.put32 (0x04);}
    	else if (fine) {data.put32 (0x03);} 
    	else data.put32 (0x02);	// Fine = 3, normal = 2, (RAW = 4 necessary for 650D)
    	
    	if (formatsToSave == 2){	// Append another data Block, indication Raw for Raw + L 
    		data.put32 (0x10);	// unknown, block length info	        	
    		data.put32 (1);     // RAW = 6, JPG = 1	        	
    		data.put32 (IMG_SIZE_L);		// S = 2, M = 1, L = 0,	
    		data.put32 (3);		// Fine = 3, normal = 2, RAW = 4
    	}
    	
    	String s = "setEOSImgQual : ";
    	if (raw) {s += (0x06);} else s +=(0x01);
    	s += size;
    	if (fine) {s += (0x03);} else s +=  (0x02);
    	Log.d(TAG, s);
		return transact0(Command.EosSetDevicePropValueEx, data);
	}
    
	
	// for debugging purpose, enter integers to be sent to camera
	public Response setEOSImgQual2 (int int1, int int2, int int3, int formatsToSave) throws PTPException{
		int blockLength = 0x0c +16* formatsToSave;
    	byte [] buff = new byte [blockLength+12]; //Total length 0x40
    	Data data = new Data (false, buff, buff.length, this);
        data.offset = 0;
        data.putHeader(buff.length, Container.BLOCK_TYPE_DATA, Command.EosSetDevicePropValueEx, 0 /*XID, dummy, will be overwritten*/);
    	data.put32 (blockLength); // Length of formats in bytes 28d = 0x1C  (int = 4 bytes)
    	data.put32 (Command.EOS_DPC_ImageQuality);
    	data.put32 (formatsToSave);	// images to store, 2 = Raw +L, 1 = single image	        	
    	data.put32 (0x10);	// unknown, block length info	        	
    	data.put32 (int1);   // RAW = 6, JPG = 1	        	
    	data.put32 (int2);	// S = 2, M = 1, L = 0	        	
    	data.put32 (int3);	// Fine = 3, normal = 2, RAW = 4
    	if (formatsToSave == 2){	// Append another data Block, indication Raw for Raw + L 
    		data.put32 (0x10);	// unknown, block length info	        	
    		data.put32 (1);   // RAW = 6, JPG = 1	        	
    		data.put32 (0);	// S = 2, M = 1, L = 0	        	
    		data.put32 (3);	// Fine = 3, normal = 2, RAW = 4
    	}
//    	data.put32 (0);
    	String s = "setEOSImgQual2: ";
    	s += int1;
    	s += int2;
    	s += int3;
    	Log.d(TAG, s);

    	return transact0(Command.EosSetDevicePropValueEx, data);
	}
    
	/*
	 *
	 * Bulb mode: http://www.gphoto.org/doc/remote/
	 * Older Canon EOS DSLR:

gphoto2 --set-config shutterspeed=bulb
gphoto2 --set-config bulb=1 --wait-event=30s --set-config bulb=0 --wait-event-and-download=2s
Newer Canon EOS DSLR:

gphoto2 --set-config shutterspeed=bulb
gphoto2 --set-config eosremoterelease=Immediate --wait-event=30s --set-config eosremoterelease=Off --wait-event-and-download=2s
	 */
	


	@Override
	public Response startBulb () throws PTPException{
//		transact0(Command.EosSetRemoteMode,null);
//		setDevicePropValue (Command.EOS_DPC_ShutterSpeed, EosEventConstants.SHUTTER_SPEED_BULB);
		
		// old command for 450D:
//		return transact0(Command.EosBulbStart,null);
		
		
		// New bulb according to usbsniffer:
		byte [] buff = new byte [0x14]; //Total length
    	Data data = new Data (false, buff, buff.length, this);
        data.offset = 0;
        data.putHeader(buff.length, Container.BLOCK_TYPE_COMMAND, Command.EosRemoteReleaseOn, 0 /*XID, dummy, will be overwritten*/);
        data.put32(1);
        data.put32(1);
        Response resp = transact0(Command.EosRemoteReleaseOn,data);
//        checkEvents();
        
        return resp;
		 // EosBulbStart not supported for 650D!
		 // Maybe for 650D bulb: libGphoto2.library.h shootingModeState: 0x07,N_("Bulb")
	}

	
	
	@Override
	public Response stopBulb () throws PTPException{
		transact0(Command.EosSetRemoteMode,null);
//		return
//				transact0(Command.EosBulbEnd,null); 
//				transact0(Command.EosGetEvent,null);
		
		
		byte [] buff = new byte [0x10]; //Total length
    	Data data = new Data (false, buff, buff.length, this);
        data.offset = 0;
        data.putHeader(buff.length, Container.BLOCK_TYPE_COMMAND, Command.EosRemoteReleaseOff, 0 /*XID, dummy, will be overwritten*/);
        data.put32(1);
        return transact0(Command.EosRemoteReleaseOff,data);
//		return  transact0(Command.EosRemoteReleaseOff,null); 
	}

    
	@Override
    public Response setShutterSpeed (int speed) throws PTPException{
        return setDevicePropValue (Command.EOS_DPC_ShutterSpeed, speed);

    }  
    
	@Override
    public Response setExposure(int exposure) throws PTPException{    	
    	return setDevicePropValue (Command.EOS_DPC_ExposureCompensation, exposure);    	
    }
        
	@Override
    public Response setISO(int value) throws PTPException{
        return  setDevicePropValue(Command.EOS_DPC_Iso ,value); 
    }
    
	@Override
    public Response setAperture(int value) throws PTPException{
        return  setDevicePropValue(Command.EOS_DPC_Aperture ,value); 
    }

    
    public Response setPictureStyle(int value) throws PTPException{
        return  setDevicePropValue(Command.EOS_DPC_PictureStyle,value); 
    }
    
    public Response setWhiteBalance(int value) throws PTPException{
        return  setDevicePropValue(Command.EOS_DPC_WhiteBalance,value); 
    }
    
    public Response setDriveMode(int value) throws PTPException{
        return  setDevicePropValue(Command.EOS_DPC_DriveMode, value); 
    }
    
    public Response setMetering(int value) throws PTPException{
        return  setDevicePropValue(Command.EOS_DPC_ExpMeterringMode, value); 
    }
    
    public Response MoveFocus(int step) throws PTPException{
        return  transact1(Command.EosDriveLens,null,step); 
    }
    
    
    // General Floating point adapters

	@Override
    public Response setShutterSpeed (double timeSeconds) throws PTPException{
    	int value = (int) (Math.round(-Math.log(timeSeconds)/Math.log(2)*4)*2 +0x38);
    	//Log.d(TAG, String.format("setShutterSpeed %.0f, value: %1d, 0x%02x", timeSeconds, value, value));
        return setShutterSpeed (value);
    }  
    
	@Override
    public Response setAperture(double apertureValue) throws PTPException{
    	int value = (int) (Math.round(Math.log(apertureValue)/Math.log(2)*4)*4 +8);
    	//Log.d(TAG, String.format("setAperture %.1f, value: %1d, 0x%02x", apertureValue, value, value));
        return  setAperture (value); 
    }
    
    // Sets ISO to 50, 100, 200... or nearest value
	@Override
    public Response setISO(double isoValue) throws PTPException{
		if (isoValue > 1) { 
			int value = (int) (Math.round(Math.log(isoValue/100.)/Math.log(2))*8 +72);
			//Log.d(TAG, String.format("SetIso %.0f, value: %1d, 0x%02x", isoValue, value, value));
			return  setISO (value);
		} else return setISO (EosEventConstants.ISO_Auto);
    }
        
	@Override
    public Response setExposure(double exposureValue) throws PTPException{   
    	if (exposureValue < -5) exposureValue = -5;
    	if (exposureValue >  5) exposureValue =  5;
    	int value = (int) (Math.round(exposureValue*8));
    	//Log.d(TAG, String.format("setExposure %.0f, value: %1d, 0x%02x", exposureValue, value, value));
    	return setExposure (value);    	
    }
    
    /**  Selects image Quality from "S" to "RAW" in 4 steps
     * 
     * @param quality  (eg "RAW", "L", "Medium", "s", LF, MN, SN, R+F)
     * @return
     * @throws PTPException
     */
	@Override
    public Response setImageQuality (String quality) throws PTPException{
    	quality = quality.toUpperCase();
    	boolean fine = (quality.contains("FINE") || quality.substring(1).startsWith("F"));
    	Response r = null;
    	if (quality.startsWith("D")) {
    		int int1  = Integer.parseInt(quality.substring(1, 2));  // RAW = 6, JPG = 1
    		int int2  = Integer.parseInt(quality.substring(2, 3));  // S = 2, M = 1, L = 0	 
    		int int3  = Integer.parseInt(quality.substring(3, 4));  // Fine = 3, normal = 2, raw = 4
    		if (int2 > 4) int2+=8;
    		r = setEOSImgQual2 (int1, int2, int3, 1);
   		Log.d(TAG, "setEOSImgQual2 s:" +int1 +" r:" +int2 +" f:" +int3); // eg.: d112 for M normal
    	}
    	if (quality.startsWith("R")) {
    		if (quality.contains("+")) {r = setEOSImgQual (0, true,  true, 2);}
    		else r = setEOSImgQual (IMG_SIZE_L, true,  true, 1); 
    	}
    	if (quality.startsWith("L")) r = setEOSImgQual (IMG_SIZE_L, false, fine || quality.length() == 1, 1);  //value = EosEventConstants.pictLFine;
    	if (quality.startsWith("M")) r = setEOSImgQual (IMG_SIZE_M, false, fine, 1);  //value = EosEventConstants.pictMFine;
    	if (quality.startsWith("S")) {
    		r = setEOSImgQual (2, false, true, 1);  //value = EosEventConstants.pictSNormal;
    		if (quality.startsWith("S1F")) {r = setEOSImgQual (IMG_SIZE_S1, false, true, 1); }
    		else if (quality.equals("S1")) {r = setEOSImgQual (IMG_SIZE_S1, false, false, 1); }
    		else if (quality.equals("S2")) {r = setEOSImgQual (IMG_SIZE_S2, false, true, 1); }
    		else if (quality.equals("S3")) {r = setEOSImgQual (IMG_SIZE_S3, false, true, 1); }
    		else r = setEOSImgQual (IMG_SIZE_S, false, fine, 1); 
    	}
//		Log.d(TAG, "Quality: " +quality +", Value: " +value);
        return  r;//setImageQuality (value); 
    }

    public Response setLiveView (boolean on) throws PTPException{
        if(on)
        { //turn on
        	return setDevicePropValueEx(Command.EOS_DPC_LiveView, 2);
        }
        else
        { //turn off
        	return setDevicePropValueEx (Command.EOS_DPC_LiveView, 0);
        }
    } 
        

}
