// Copyright 2000 by David Brownell <dbrownell@users.sourceforge.net>
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
package com.android.ptplib.nikon;


import java.util.Vector;

import android.hardware.usb.UsbDevice;
import android.hardware.usb.UsbDeviceConnection;
import android.util.Log;
import com.android.ptplib.BaselineInitiator;
import com.android.ptplib.CameraFunctions;
import com.android.ptplib.CameraManager;
import com.android.ptplib.Command;
import com.android.ptplib.Container;
import com.android.ptplib.Data;
import com.android.ptplib.DevicePropDesc;
import com.android.ptplib.PTPException;
import com.android.ptplib.Response;

/**
 * This supports all standardized PTP-over-USB operations, including
 * operations (and modes) that are optional for all responders.
 * Filtering operations invoked on this class may be done on the device,
 * or may be emulated on the client side.
 * At this time, not all standardized operations are supported.
 *
 * @author MagicAsh and flying_galactic, based on JAVA code by David Brownell
 *
 * This class has been reworked by flying_galactic and MagicAsh order to make it 
 * compatible with Android.
 *
 */
public class NikonInitiator extends BaselineInitiator implements CameraFunctions {
	
	public static final int NIKON_VID = 0x04b0; //http://www.linux-usb.org/usb.ids

	
	public static boolean eventListenerRunning = false;
    /**
     * This is essentially a class driver, following Annex D of
     * the PTP specification.
     */
    public NikonInitiator(UsbDevice dev, UsbDeviceConnection connection) throws PTPException {
        super(dev, connection);
        /*
        if (thread != null) eventListenerRunning = false;
        thread = new Thread (this);
        thread.start();
        */
    }

    /**
     * Fills out the provided device property description.
     *
     * @param propcode code identifying the property of interest
     * @param desc description to be filled; it may be a subtype
     *	associated with with domain-specific methods
     * @return response code
     */
    public int getDevicePropDesc(int propcode, DevicePropDesc desc)
            throws PTPException {
        return transact1(Command.GetDevicePropDesc, desc, propcode).getCode();
    }




    /**
     * Starts the capture of one (or more) new
     * data objects, according to current device properties.
     * The capture will complete without issuing further commands.
     *
     * @see #initiateOpenCapture
     *
     * @param storageId Where to store the object(s), or zero to
     *	let the device choose.
     * @param formatCode Type of object(s) to capture, or zero to
     *	use the device default.
     *
     * @return status code indicating whether capture started;
     *	CaptureComplete events provide capture status, and
     *	ObjectAdded events provide per-object status.
     */
    public Response initiateCapture(int storageId, int formatCode)
            throws PTPException {
    	Response resp = null;
        if (!info.supportsOperation(Command.InitiateCapture)) {
        	Log.d(TAG, "The camera does not support Nikibon capture");
            throw new PTPException("The camera does not support Nikon capture");
        }

        int ret;

        resp = transact0(Command.InitiateCapture, null);
        ret = resp.getCode();
        Log.d(TAG, "  NK_OC_Capture Response code: 0x" +Integer.toHexString(ret) +"  OK: " +(ret == Response.OK));
        if (ret != Response.OK) {
            String msg = "NK_OC_Capture  Capture failed to release: Unknown error "
                    + ret
                    + " , please report.";
            Log.d(TAG, msg);
            throw new PTPException(msg, ret);
        }

      try { Thread.sleep(100);  } catch (InterruptedException e) {e.printStackTrace();}


      return resp;
    }

    /**
     * Retrieves a chunk of the object identified by the given object id.
     *
     * @param oid object id
     * @param offset the offset to start from
     * @param size the number of bytes to transfer
     * @param data the Data object receiving the object
     *
     * @throws PTPException in case of errors
     */
    public void getPartialObject(int oid, int offset, int size, Data data)
    throws PTPException {
        Response ret =
            transact3(Command.EosGetPartialObject, data, oid, offset, size);

        if (ret.getCode() != Response.OK) {
            throw new PTPException("Error reading new object", ret.getCode());
        }
    }

    public void transferComplete(int oid)
    throws PTPException {
        Response ret =
            transact1(Command.EosTransferComplete, null, oid);

        if (ret.getCode() != Response.OK) {
            throw new PTPException("Error reading new object", ret.getCode());
        }
    }

    public Response setDevicePropValue (int property, int value) throws PTPException{
    	return setDevicePropValueNikon (property, value);
    }
    
    //Methods below Added, getDevicePropDesc should work too
    
    public void writeExtraData(Command command, Data data, int timeout)
    {
    	int lenC = mConnection.bulkTransfer(epOut, command.data , command.length , timeout);

    	if ((command.length % epOut.getMaxPacketSize()) == 0) {
    		lenC = mConnection.bulkTransfer(epOut, command.data, 0, timeout);
    	}
    	//////////////////////////////////// 
    	int opcode = command.getCode();
    	data.offset = 0;
    	data.putHeader(data.getLength(), 2 , opcode, command.getXID());
    	byte[] bytes = data.getData();

    	lenC = mConnection.bulkTransfer(epOut, data.getData() , data.length , timeout);

    	if ((data.length % epOut.getMaxPacketSize()) == 0) {
    		lenC = mConnection.bulkTransfer(epOut, bytes , 0, timeout);
    	}
    }

    public Response setDevicePropValueNikon (int property, int value) throws PTPException{        
        byte [] buff = new byte [0x14];
        Data data = new Data (false, buff, buff.length, this);
        data.offset = 0;
        data.putHeader(buff.length, Container.BLOCK_TYPE_DATA, Command.SetDevicePropValue, 0 );
        data.put32 (value);
        Command command = new Command(Command.SetDevicePropValue, session, property);
        writeExtraData(command, data, DEFAULT_TIMEOUT);
        
        // read response
        byte buf[] = new byte[inMaxPS];
        int len = mConnection.bulkTransfer(epIn, buf ,inMaxPS , DEFAULT_TIMEOUT);
        Response response = new Response(buf, len, this);
        return response;
    }
    
    public Response setDevicePropValueNikon (int property, String value) throws PTPException{        
        byte [] buff = new byte [0x14 +value.length()*2];
        Data data = new Data (false, buff, buff.length, this);
        data.offset = 0;
        data.putHeader(buff.length, Container.BLOCK_TYPE_DATA, Command.SetDevicePropValue, 0 );
        data.putString(value);    
        Command command = new Command(Command.SetDevicePropValue, session, property);
        writeExtraData(command, data, DEFAULT_TIMEOUT);
        
        // read response
        byte buf[] = new byte[inMaxPS];
        int len = mConnection.bulkTransfer(epIn, buf ,inMaxPS , DEFAULT_TIMEOUT);
        Response response = new Response(buf, len, this);
        return response;
    }
    
    @Override
    public Response getDevicePropValue(int property, DevicePropDesc desc)
    		throws PTPException {
		Command command = new Command(Command.GetDevicePropDesc, session, property);
		mConnection.bulkTransfer(epOut, command.data, command.length, BaselineInitiator.DEFAULT_TIMEOUT);    
		byte buf[] = read(BaselineInitiator.DEFAULT_TIMEOUT);

		desc.data = buf;
		desc.length = info.getLength();
		desc.parse(); // crashes on 2nd call without reading response!
		
		// do not forget, read response otherwise may crash!		
        buf = new byte[inMaxPS];
        int len = mConnection.bulkTransfer(epIn, buf ,inMaxPS , DEFAULT_TIMEOUT);
        return new Response(buf, len, this);
	}
    
	public DevicePropDesc getPropValue (int property) { // for NIKON only!
		Command command = new Command(Command.GetDevicePropDesc, session, property);
		mConnection.bulkTransfer(epOut, command.data, command.length, BaselineInitiator.DEFAULT_TIMEOUT);    
		byte buf[] = read(BaselineInitiator.DEFAULT_TIMEOUT);

		DevicePropDesc info = new DevicePropDesc (this);
		info.data = buf;
		info.length = info.getLength();
		info.parse(); // crashes on 2nd call without reading response!

		// do not forget, read response otherwise may crash!		
        buf  = new byte[inMaxPS];
        int len = mConnection.bulkTransfer(epIn, buf ,inMaxPS , DEFAULT_TIMEOUT);
        Response response = new Response(buf, len, this);
		return info;
	}

    
    
    public Response MoveFocus(int step) throws PTPException{

        return  null;//transact1(Command.EosDriveLens,null,step); 

    }
    
    public Response setExposure(int exposure) throws PTPException{
    	
    	return setDevicePropValueNikon(NikonEventConstants.PTP_DPC_ExposureBiasCompensation, exposure);
    	
    }
    
    public Response setISO(int value) throws PTPException{  //Caution: NIKON only accepts correct values!

        return  setDevicePropValueNikon(NikonEventConstants.PTP_DPC_ExposureIndex ,value); 
    }
    
    public Response setAperture(int value) throws PTPException{

        return  setDevicePropValueNikon(NikonEventConstants.PTP_DPC_FNumber ,value); 
    }
    
    
    public Response setPictureStyle(int value) throws PTPException{

        return null; //return  setDevicePropValueNikon(Command.EOS_DPC_PictureStyle,value); 
    }
    
    public Response setWhiteBalance(int value) throws PTPException{

        return  setDevicePropValueNikon(NikonEventConstants.PTP_DPC_WhiteBalance,value); 
    }
    
    public Response setDriveMode(int value) throws PTPException{

        return null;//return  setDevicePropValueNikon(Command.EOS_DPC_DriveMode, value); 
    }
    
    public Response setMetering(int value) throws PTPException{

        return  null;//setDevicePropValueNikon(NikonEventConstants.PTP_DPC_FocusMeteringMode, value); 
    }
    
    public Response setImageQuality(int value) throws PTPException{

        return  null;//setDevicePropValueNikon(Command.EOS_DPC_ExpMeterringMode, value); 
    }
    
    public Response setShutterSpeed (int speed) throws PTPException{
        return setDevicePropValueNikon (NikonEventConstants.PTP_DPC_ExposureTime, speed);

    }   
    
    // Floating point adapters

	@Override
    public Response setShutterSpeed (double timeSeconds) throws PTPException{
    	int value = (int) (/*Math.round*/(timeSeconds*10000.));
    	Log.d(TAG, String.format("setShutterSpeed %.0f, value: %1d, 0x%02x", timeSeconds, value, value));
        return setShutterSpeed (value);
    }  
    
	@Override
    public Response setAperture(double apertureValue) throws PTPException{
    	int value = (int) (Math.round(apertureValue*100));
    	Log.d(TAG, String.format("setAperture %.1f, value: %1d, 0x%02x", apertureValue, value, value));
        return  setAperture (value); 
    }
    
    // Sets ISO to 50, 100, 200... or nearest value. Caution: NIKON only accepts correct values!
	@Override
    public Response setISO(double isoValue) throws PTPException{
		if (isoValue > 1) { 
			int value = (int) isoValue;
			Log.d(TAG, String.format("SetIso %.0f, value: %1d, 0x%02x", isoValue, value, value));
			return  setISO (value);
		} else return //setDevicePropValueNikon(NikonEventConstants.NK_DPC_ISOAuto ,0);  
				setISO (NikonEventConstants.NK_DPC_ISOAuto);
    }
        
	@Override
    public Response setExposure(double exposureValue) throws PTPException{   
    	if (exposureValue < -5) exposureValue = -5;
    	if (exposureValue >  5) exposureValue =  5;
    	int value = (int) (Math.round(Math.round(exposureValue*3)*333.3333)) & 0xFFFF; //Does not work properly
    	Log.d(TAG, String.format("setExposure %.2f, value: %1d, 0x%02x", exposureValue, value, value));
    	Response r = setExposure (value);
    	if (r.getCode() == Response.InvalidDevicePropValue) {
    		value = (int) (Math.round(Math.round(exposureValue*2)*500.)) & 0xFFFF;
        	Log.d(TAG, String.format("setExposure %.2f, value: %1d, 0x%02x", exposureValue, value, value));
    		r = setExposure (value);
    	}
    	return r;
    }

	private byte[] read(int timeout)
	{
		Log.d(TAG,"Reading data");
		byte data[] = new byte[inMaxPS]; 
		int lengthOfBytes = mConnection.bulkTransfer(epIn, data, inMaxPS, timeout);	  
		return data;	  
	}

	
    
    /**  Selects image Quality from "S" to "RAW" 
     * 
     * @param quality  (eg "RAW", "L", "Medium", "s", LF, MN, SB, R+B)
     * @return
     * @throws PTPException
     */
	
	@Override
    public Response setImageQuality (String quality) throws PTPException{
    	quality = quality.toUpperCase();
    	int compression = NikonEventConstants.NK_Compression_BASIC;
    	if (quality.contains("FINE") || quality.substring(1).startsWith("F"))
    		compression = NikonEventConstants.NK_Compression_FINE;
    	if (quality.contains("NORM") || quality.substring(1).startsWith("N"))
    		compression = NikonEventConstants.NK_Compression_NORM;
    		
    		
    	Response r = null;
    	if (quality.startsWith("R")) {
    		r = setImageSize ("L"); // not needed, will we overridden by any RAW compression  
    		if (quality.contains("+")) {
    			compression = NikonEventConstants.NK_Compression_RAW_BASIC;
    		} else compression = NikonEventConstants.NK_Compression_RAW_B;
    		r = setDevicePropValueNikon(NikonEventConstants.PTP_DPC_CompressionSetting, compression);
    	}
    	if (quality.startsWith("L")) {
    		r = setImageSize ("L");
    		if (quality.length() == 1) compression = NikonEventConstants.NK_Compression_FINE;
    		r = setDevicePropValueNikon(NikonEventConstants.PTP_DPC_CompressionSetting, compression);
    	}
    	if (quality.startsWith("M")) {
    		r = setImageSize ("M");
    		if (quality.length() == 1) compression = NikonEventConstants.NK_Compression_NORM;
    		r = setDevicePropValueNikon(NikonEventConstants.PTP_DPC_CompressionSetting, compression);
    	}
    	if (quality.startsWith("S")) {
    		r = setImageSize ("S");
    		r = setDevicePropValueNikon(NikonEventConstants.PTP_DPC_CompressionSetting, compression);
    	}
        return  r;//setImageQuality (value); 
    }

    /** Sets image size due to a symbolic value like ("S", "M", "L")
     * 
     * @param  symbolic "S", "M", "L"
     * @return response
     * @throws PTPException
     */
    public Response setImageSize (String symbolic) throws PTPException {
    	// reading possible image resolutions
    	DevicePropDesc info = new DevicePropDesc (this);
    	info = getPropValue (NikonEventConstants.PTP_DPC_ImageSize);
    	
    	Vector v = (Vector) info.constraints;
    	boolean ascending = true;
    	if (v.size () > 1 && CameraManager.getNumeric((String) v.elementAt (0)) > CameraManager.getNumeric((String) (v.elementAt(v.size ()-1))))
    		ascending = false;

    	symbolic = symbolic.toUpperCase().trim().substring(0,1);
    	String resolution; // will contain a string like "3872x2592"
    	
    	if (symbolic.equals("M")) {resolution = (String) v.elementAt ((int) (Math.round(v.size ()-0.1)/2));}
    	else if (symbolic.equals("S")) {
    		if (ascending) {resolution = (String) v.elementAt (0);}
    		else resolution = (String) v.elementAt (v.size ()-1);
    	} else {
    		if (ascending) {resolution = (String) v.elementAt (v.size ()-1);}
    		else resolution = (String) v.elementAt (0);
    	}
    	Log.d(TAG, "set imageSize: " +symbolic + "  Resolution: " +resolution);
    	return setDevicePropValueNikon(NikonEventConstants.PTP_DPC_ImageSize, resolution);
    }


    @Override    
    public void run() {

    }

    
    
}
